import datetime

CURRENT_TIME, CURRENT_TIME_STR = datetime.datetime(2032, 5, 23, 0, 0, 0), "2032-05-23T00:00:00"
FUTURE_TIME = datetime.datetime(2042, 5, 23, 0, 0, 0)
SOME_TIME, SOME_TIME_STR = datetime.datetime(2022, 5, 10, 17, 50, 32), "2022-05-10T17:50:32"
ANOTHER_TIME, ANOTHER_TIME_STR = datetime.datetime(2022, 5, 11, 17, 50, 32), "2022-05-11T17:50:32"
HOST = "http://test-api"
ONE_TO_TEN = list(range(1, 10))
